unit qiGroup;
{
Copyright 1996 by Justin Turberville
EMail: JustinT@cyberjoe.co.za
}
interface

uses
  WinTypes, WinProcs, Classes, Graphics, Forms, Controls, Buttons,
  StdCtrls, ExtCtrls, DB, DBTables, SysUtils, qiField, qiQuery;

type
  TQIGroupDialog = class(TComponent)
  private
    FAbout: TQIAboutProperty;
    FCaption:     string;
    FFont:        TFont;
    FAddTableName: Boolean;
    FMaxItems:    Integer;
    FModalResult: TModalResult;
    FQueryInfo:   TQIQueryInfo;
    FShowGlyphs:  Boolean;
    FSortFields:  Boolean;
    FSelectedOnly: Boolean;
    procedure SetFont(Value: TFont);
    procedure SetQueryInfo(Value: TQIQueryInfo);
    procedure SetMaxItems(Value: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function Execute: Boolean;
    property ModalResult: TModalResult read FModalResult write FModalResult;
  published
    property About: TQIAboutProperty read FAbout write FAbout stored False;
    property Caption: string read FCaption write FCaption;
    property Font: TFont read FFont write SetFont;
    property QueryInfo: TQIQueryInfo read FQueryInfo write SetQueryInfo;
    property ShowGlyphs: Boolean read FShowGlyphs write FShowGlyphs default False;
    property MaxItems: Integer read FMaxItems write SetMaxItems default 10;
    property SortFields: Boolean read FSortFields write FSortFields default False;
    property AddTableName: Boolean read FAddTableName write FAddTableName default False;
    property SelectedOnly: Boolean read FSelectedOnly write FSelectedOnly default False;
  end;

  TQIGroupForm = class(TForm)
    FieldLabel: TLabel;
    FieldList: TListBox;
    InclBtn: TSpeedButton;
    ExclBtn: TSpeedButton;
    SortLabel: TLabel;
    EditList: TListBox;
    ClearBtn: TSpeedButton;
    UpBtn: TSpeedButton;
    DownBtn: TSpeedButton;
    OkBtn: TBitBtn;
    CancelBtn: TBitBtn;
    Bevel1: TBevel;
    procedure FormShow(Sender: TObject);
    procedure ListDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure ListDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure InclBtnClick(Sender: TObject);
    procedure ExclBtnClick(Sender: TObject);
    procedure FieldListDblClick(Sender: TObject);
    procedure EditListDblClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure UpBtnClick(Sender: TObject);
    procedure DownBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure EditListKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ClearBtnClick(Sender: TObject);
    procedure ListClick(Sender: TObject);
  private
    procedure UpdateBtns;
  public
    QueryInfo: TQIQueryInfo;
    MaxItems: Integer;
    AddTableName: Boolean;
    SelectedOnly: Boolean;
  end;

implementation

{$R *.DFM}

type
  TEditObj = class(TObject)
    StoreItem, EditItem: TGroupItem;
    constructor Create(AItem: TObject);
    destructor Destroy; override;
    procedure AssignEdit;
    function ListName(ListTable: Boolean): string;
  end;

{TQIGroupDialog}

constructor TQIGroupDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCaption := 'Define Grouping';
  FFont := TFont.Create;
  FMaxItems := 10;
end;

destructor TQIGroupDialog.Destroy;
begin
  FFont.Free;
  inherited Destroy;
end;

procedure TQIGroupDialog.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FQueryInfo) then
    FQueryInfo := nil;
end;

function TQIGroupDialog.Execute: Boolean;
begin
  Result := False;
  if FQueryInfo = nil then
    raise EComponentError.Create(sNoQueryInfo);
  with TQIGroupForm.Create(Self) do
  try
    Caption := FCaption;
    Font.Assign(FFont);
    QueryInfo := FQueryInfo;
    MaxItems := FMaxItems;
    AddTableName := FAddTableName;
    FieldList.Sorted := FSortFields;
    SelectedOnly := FSelectedOnly;
    if FShowGlyphs then begin
      OKBtn.Kind := bkOK;

      CancelBtn.Kind := bkCancel;

    end;

    FModalResult := ShowModal;
  finally

    Free;

    Result := FModalResult = mrOK;
  end;
end;

procedure TQIGroupDialog.SetFont(Value: TFont);
begin
  if Value <> nil then
    FFont.Assign(Value);
end;

procedure TQIGroupDialog.SetQueryInfo(Value: TQIQueryInfo);
begin
  FQueryInfo := Value;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TQIGroupDialog.SetMaxItems(Value: Integer);
begin
  if Value < 1 then FMaxItems := 1
  else if Value > 20 then FMaxItems := 20
  else FMaxItems := Value;
end;

{TEditObj}

constructor TEditObj.Create(AItem: TObject);
begin
  inherited Create;
  EditItem := TGroupItem.Create;
  if AItem is TGroupItem then begin         {an existing item -}
    StoreItem := TGroupItem(AItem);
    StoreItem.EditFlag := efOld;
    EditItem.Assign(StoreItem);
  end
  else begin                                 {a new item -}
    StoreItem := TGroupItem.Create;
    StoreItem.EditFlag := efNew;
    if AItem is TFieldItem then
      EditItem.Field := TFieldItem(AItem);
  end;
end;

destructor TEditObj.Destroy;
begin
  EditItem.Free;
  if StoreItem.EditFlag = efNew then   {a new item}
    StoreItem.Free else                {so destroy (should'nt be in Grouping)}
    StoreItem.EditFlag := efDelete;    {else mark as to delete if edit OK'd}
  inherited Destroy;                             {- is in Grouping list}
end;

procedure TEditObj.AssignEdit;
begin
  StoreItem.Assign(EditItem);
  StoreItem.EditFlag := efOld;
end;

function TEditObj.ListName(ListTable: Boolean): string;
begin
  with EditItem do
    if (Field = nil) or (Field.UserName = '') then
      Result := sUnknown else
      Result := Field.UserNameStr(ListTable);
end;

{TQIGroupForm}

procedure TQIGroupForm.FormShow(Sender: TObject);
var
  i: Integer;
  b: Boolean;
  e: TEditObj;
begin
  with QueryInfo do begin
    for i := 0 to Grouping.Count-1 do begin        {add existing groupers -}
      if SelectedOnly then                {if list only selected fields}
        b := Selections.ByField(Grouping[i].Field) <> nil else
        b := True;
      if b then begin
        e := TEditObj.Create(Grouping[i]);
        EditList.Items.AddObject(e.ListName(AddTableName), e);
      end;
    end;                                            {add fields -}
    for i := 0 to Fields.Count-1 do begin  {add if not in grouping & can be grouped -}
      if SelectedOnly then                {if list only selected fields}
        b := Selections.ByField(Fields[i]) <> nil else  {then check if selected}
        b := True;
      if b and (Grouping.ByField(Fields[i]) = nil) and
         (fuSelect in Fields[i].Usage) and
         (tuQuery in Fields[i].Table.Usage) then
        FieldList.Items.AddObject(
          Fields[i].UserNameStr(AddTableName), Fields[i]);
    end;
  end;
  if EditList.Items.Count > 0 then          {select first grouper -}
    EditList.ItemIndex := 0;
  UpdateBtns;
end;

procedure TQIGroupForm.ListDragOver(Sender, Source: TObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
begin
  if Source = EditList then
    if Sender = EditList then
      Accept := EditList.Items.Count > 0 else           {moving grouping item}
      Accept := ExclBtn.Enabled                         {excluding field}
  else if Source = FieldList  then
    Accept := (Source <> Sender) and InclBtn.Enabled;   {including field}
end;

procedure TQIGroupForm.ListDragDrop(Sender, Source: TObject; X, Y: Integer);

  procedure Include(I: Integer);
  var
    e: TEditObj;
  begin
    e := TEditObj.Create(FieldList.Items.Objects[I]);
    with EditList do
      ItemIndex := Items.AddObject(e.ListName(AddTableName), e);
    FieldList.Items.Delete(I);
  end;

  procedure Exclude(I: Integer);
  var
    e: TEditObj;
  begin
    e := TEditObj(EditList.Items.Objects[I]);
    with FieldList, e.EditItem do    {add field to available list -}
      Selected[Items.AddObject(Field.UserNameStr(AddTableName), Field)] := True;
    EditList.Items.Delete(I);
    e.Free;
  end;

var
  i: Integer;
begin  {ListDragDrop}
  if Source = Sender then               {a EditList drag move -}
    with EditList do begin
      i := ItemAtPos(Point(X, Y), False);
      if i >= Items.Count then Dec(i);
      Items.Move(ItemIndex, i);
      ItemIndex := i;
    end
  else if Source = FieldList then begin      {including a field -}
    i := 0;
    while i < FieldList.Items.Count do
      if FieldList.Selected[i] then
        if EditList.Items.Count >= MaxItems then Break
        else Include(i)
      else Inc(i)
  end
  else if Source = EditList then            {excluding a field -}
    with EditList do begin
      i := ItemIndex;
      Exclude(ItemIndex);
      if i > Items.Count-1 then Dec(i);
      if i > -1 then ItemIndex := i;
    end;
  UpdateBtns;
end;

procedure TQIGroupForm.InclBtnClick(Sender: TObject);
begin
  ListDragDrop(EditList, FieldList, 0, 0);
end;

procedure TQIGroupForm.ExclBtnClick(Sender: TObject);
begin
  ListDragDrop(FieldList, EditList, 0, 0);
end;

procedure TQIGroupForm.FieldListDblClick(Sender: TObject);
begin
 if InclBtn.Enabled then
   ListDragDrop(EditList, FieldList, 0, 0);
end;

procedure TQIGroupForm.EditListDblClick(Sender: TObject);
begin
 if ExclBtn.Enabled then
   ListDragDrop(FieldList, EditList, 0, 0);
end;

procedure TQIGroupForm.UpdateBtns;
begin
  with EditList do begin
    InclBtn.Enabled := (FieldList.SelCount > 0) and (Items.Count < MaxItems);
    ExclBtn.Enabled := ItemIndex > -1;
    UpBtn.Enabled   := ItemIndex > 0;
    DownBtn.Enabled := ExclBtn.Enabled and (ItemIndex < Items.Count-1);
    ClearBtn.Enabled := Items.Count > 0;
  end;
end;

procedure TQIGroupForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_INSERT: if InclBtn.Enabled then InclBtnClick(nil);
    VK_DELETE: if ExclBtn.Enabled then ExclBtnClick(nil);
  end;
end;

procedure TQIGroupForm.UpBtnClick(Sender: TObject);
begin
  with EditList do begin
    Items.Move(ItemIndex -1, ItemIndex);
    if ItemIndex < TopIndex then          {if above listbox top}
      TopIndex := TopIndex -1;            {then scroll up}
  end;
  UpdateBtns;
end;

procedure TQIGroupForm.DownBtnClick(Sender: TObject);
begin
  with EditList do begin
    Items.Move(ItemIndex +1, ItemIndex);
    if ItemIndex >= TopIndex + (Height div ItemHeight) then   {if below bottom}
      TopIndex := TopIndex +1;                                {then scroll down}
  end;
  UpdateBtns;
end;

procedure TQIGroupForm.FormDestroy(Sender: TObject);
var
  i: Integer;
begin
  if ModalResult = mrOK then
    with QueryInfo, Grouping do begin
      while Count > 0 do          {clear to restore in edited order (free if}
        Remove(Items[0], Items[0].EditFlag = efDelete);   { deleted in edit)}
      for i := 0 to EditList.Items.Count-1 do     {add edited grouping -}
        with TEditObj(EditList.Items.Objects[i]) do begin
          AssignEdit;                {assign edited properties to group item}
          Add(StoreItem);            {add grouper to Grouping list}
        end;
      ChangeNotify(Owner, Grouping);        {fire event notifier}
    end;
  with EditList.Items do     {destroy edit objects -}
    for i := 0 to Count-1 do
      Objects[i].Free;
end;

procedure TQIGroupForm.EditListKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if ssCtrl in Shift then
    case Key of
      VK_UP, VK_LEFT: begin
        if UpBtn.Enabled then UpBtnClick(nil);
        Key := 0;
      end;
      VK_DOWN, VK_RIGHT: begin
        if DownBtn.Enabled then DownBtnClick(nil);
        Key := 0;
      end;
      VK_HOME:
        if UpBtn.Enabled then begin
          with EditList do Items.Move(ItemIndex, 0);
          UpdateBtns;
        end;
      VK_END:
        if DownBtn.Enabled then begin
          with EditList do Items.Move(ItemIndex, Items.Count-1);
          UpdateBtns;
        end;
    end;
end;

procedure TQIGroupForm.ClearBtnClick(Sender: TObject);
begin
  while EditList.Items.Count > 0 do begin
    EditList.ItemIndex := 0;
    ExclBtnClick(nil);
  end;
end;

procedure TQIGroupForm.ListClick(Sender: TObject);
begin
  UpdateBtns;
end;

end.
